@@ -1,5 +1,5 @@
 /*
- * Copyright 2017 MongoDB, Inc.
+ * Copyright 2017-present MongoDB, Inc.
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -36,7 +36,7 @@ void php_phongo_maxkey_interface_init_ce(INIT_FUNC_ARGS) /* {{{ */
 	zend_class_entry ce;
 
 	INIT_NS_CLASS_ENTRY(ce, "MongoDB\\BSON", "MaxKeyInterface", php_phongo_maxkey_interface_me);
-	php_phongo_maxkey_interface_ce = zend_register_internal_interface(&ce TSRMLS_CC);
+	php_phongo_maxkey_interface_ce = zend_register_internal_interface(&ce);
 } /* }}} */
 
 /*
