@@ -1,5 +1,5 @@
 /*
- * Copyright 2017 MongoDB, Inc.
+ * Copyright 2017-present MongoDB, Inc.
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -38,26 +38,25 @@ zend_class_entry* php_phongo_session_ce;
 #define SESSION_CHECK_LIVELINESS(i, m)                                  \
 	if (!(i)->client_session) {                                         \
 		phongo_throw_exception(                                         \
-			PHONGO_ERROR_LOGIC TSRMLS_CC,                               \
+			PHONGO_ERROR_LOGIC,                                         \
 			"Cannot call '%s', as the session has already been ended.", \
 			(m));                                                       \
 		return;                                                         \
 	}
 
-static bool php_phongo_session_get_timestamp_parts(zval* obj, uint32_t* timestamp, uint32_t* increment TSRMLS_DC)
+static bool php_phongo_session_get_timestamp_parts(zval* obj, uint32_t* timestamp, uint32_t* increment)
 {
-	bool retval = false;
-#if PHP_VERSION_ID >= 70000
+	bool retval     = false;
 	zval ztimestamp = ZVAL_STATIC_INIT;
 	zval zincrement = ZVAL_STATIC_INIT;
 
-	zend_call_method_with_0_params(obj, NULL, NULL, "getTimestamp", &ztimestamp);
+	zend_call_method_with_0_params(PHONGO_COMPAT_OBJ_P(obj), NULL, NULL, "getTimestamp", &ztimestamp);
 
 	if (Z_ISUNDEF(ztimestamp) || EG(exception)) {
 		goto cleanup;
 	}
 
-	zend_call_method_with_0_params(obj, NULL, NULL, "getIncrement", &zincrement);
+	zend_call_method_with_0_params(PHONGO_COMPAT_OBJ_P(obj), NULL, NULL, "getIncrement", &zincrement);
 
 	if (Z_ISUNDEF(zincrement) || EG(exception)) {
 		goto cleanup;
@@ -65,25 +64,6 @@ static bool php_phongo_session_get_timestamp_parts(zval* obj, uint32_t* timestam
 
 	*timestamp = Z_LVAL(ztimestamp);
 	*increment = Z_LVAL(zincrement);
-#else
-	zval* ztimestamp = NULL;
-	zval* zincrement = NULL;
-
-	zend_call_method_with_0_params(&obj, NULL, NULL, "getTimestamp", &ztimestamp);
-
-	if (Z_ISUNDEF(ztimestamp) || EG(exception)) {
-		goto cleanup;
-	}
-
-	zend_call_method_with_0_params(&obj, NULL, NULL, "getIncrement", &zincrement);
-
-	if (Z_ISUNDEF(zincrement) || EG(exception)) {
-		goto cleanup;
-	}
-
-	*timestamp = Z_LVAL_P(ztimestamp);
-	*increment = Z_LVAL_P(zincrement);
-#endif
 
 	retval = true;
 
@@ -99,7 +79,7 @@ cleanup:
 	return retval;
 }
 
-static const char* php_phongo_get_transaction_state_string(mongoc_transaction_state_t state TSRMLS_DC)
+static const char* php_phongo_get_transaction_state_string(mongoc_transaction_state_t state)
 {
 	switch (state) {
 		case MONGOC_TRANSACTION_NONE:
@@ -113,15 +93,71 @@ static const char* php_phongo_get_transaction_state_string(mongoc_transaction_st
 		case MONGOC_TRANSACTION_ABORTED:
 			return PHONGO_TRANSACTION_ABORTED;
 		default:
-			phongo_throw_exception(PHONGO_ERROR_INVALID_ARGUMENT TSRMLS_CC, "Invalid transaction state %d given", (int) state);
+			phongo_throw_exception(PHONGO_ERROR_INVALID_ARGUMENT, "Invalid transaction state %d given", (int) state);
 			return NULL;
 	}
 }
 
+static void php_phongo_transaction_options_to_zval(mongoc_client_session_t* cs, zval* retval)
+{
+	mongoc_transaction_opt_t*     opts;
+	int64_t                       max_commit_time_ms;
+	const mongoc_read_concern_t*  read_concern;
+	const mongoc_read_prefs_t*    read_preference;
+	const mongoc_write_concern_t* write_concern;
+
+	if (!cs) {
+		ZVAL_NULL(retval);
+		return;
+	}
+
+	opts = mongoc_session_opts_get_transaction_opts(cs);
+
+	if (!opts) {
+		ZVAL_NULL(retval);
+		return;
+	}
+
+	max_commit_time_ms = mongoc_transaction_opts_get_max_commit_time_ms(opts);
+	read_concern       = mongoc_transaction_opts_get_read_concern(opts);
+	read_preference    = mongoc_transaction_opts_get_read_prefs(opts);
+	write_concern      = mongoc_transaction_opts_get_write_concern(opts);
+
+	array_init_size(retval, 4);
+
+	if (max_commit_time_ms) {
+		ADD_ASSOC_LONG_EX(retval, "maxCommitTimeMS", max_commit_time_ms);
+	}
+
+	if (!mongoc_read_concern_is_default(read_concern)) {
+		zval zread_concern;
+
+		phongo_readconcern_init(&zread_concern, read_concern);
+		ADD_ASSOC_ZVAL_EX(retval, "readConcern", &zread_concern);
+	}
+
+	if (read_preference) {
+		zval zread_preference;
+
+		phongo_readpreference_init(&zread_preference, read_preference);
+		ADD_ASSOC_ZVAL_EX(retval, "readPreference", &zread_preference);
+	}
+
+	if (!mongoc_write_concern_is_default(write_concern)) {
+		zval zwrite_concern;
+
+		phongo_writeconcern_init(&zwrite_concern, write_concern);
+		ADD_ASSOC_ZVAL_EX(retval, "writeConcern", &zwrite_concern);
+	}
+
+	mongoc_transaction_opts_destroy(opts);
+}
+
 /* {{{ proto void MongoDB\Driver\Session::advanceClusterTime(array|object $clusterTime)
    Advances the cluster time for this Session */
 static PHP_METHOD(Session, advanceClusterTime)
 {
+	zend_error_handling   error_handling;
 	php_phongo_session_t* intern;
 	zval*                 zcluster_time;
 	bson_t                cluster_time = BSON_INITIALIZER;
@@ -129,11 +165,14 @@ static PHP_METHOD(Session, advanceClusterTime)
 	intern = Z_SESSION_OBJ_P(getThis());
 	SESSION_CHECK_LIVELINESS(intern, "advanceClusterTime")
 
-	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "A", &zcluster_time) == FAILURE) {
+	zend_replace_error_handling(EH_THROW, phongo_exception_from_phongo_domain(PHONGO_ERROR_INVALID_ARGUMENT), &error_handling);
+	if (zend_parse_parameters(ZEND_NUM_ARGS(), "A", &zcluster_time) == FAILURE) {
+		zend_restore_error_handling(&error_handling);
 		return;
 	}
+	zend_restore_error_handling(&error_handling);
 
-	php_phongo_zval_to_bson(zcluster_time, PHONGO_BSON_NONE, &cluster_time, NULL TSRMLS_CC);
+	php_phongo_zval_to_bson(zcluster_time, PHONGO_BSON_NONE, &cluster_time, NULL);
 
 	/* An exception may be thrown during BSON conversion */
 	if (EG(exception)) {
@@ -150,6 +189,7 @@ cleanup:
    Advances the operation time for this Session */
 static PHP_METHOD(Session, advanceOperationTime)
 {
+	zend_error_handling   error_handling;
 	php_phongo_session_t* intern;
 	zval*                 ztimestamp;
 	uint32_t              timestamp = 0;
@@ -158,11 +198,14 @@ static PHP_METHOD(Session, advanceOperationTime)
 	intern = Z_SESSION_OBJ_P(getThis());
 	SESSION_CHECK_LIVELINESS(intern, "advanceOperationTime")
 
-	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "O", &ztimestamp, php_phongo_timestamp_interface_ce) == FAILURE) {
+	zend_replace_error_handling(EH_THROW, phongo_exception_from_phongo_domain(PHONGO_ERROR_INVALID_ARGUMENT), &error_handling);
+	if (zend_parse_parameters(ZEND_NUM_ARGS(), "O", &ztimestamp, php_phongo_timestamp_interface_ce) == FAILURE) {
+		zend_restore_error_handling(&error_handling);
 		return;
 	}
+	zend_restore_error_handling(&error_handling);
 
-	if (!php_phongo_session_get_timestamp_parts(ztimestamp, &timestamp, &increment TSRMLS_CC)) {
+	if (!php_phongo_session_get_timestamp_parts(ztimestamp, &timestamp, &increment)) {
 		return;
 	}
 
@@ -173,6 +216,7 @@ static PHP_METHOD(Session, advanceOperationTime)
    Returns the cluster time for this Session */
 static PHP_METHOD(Session, getClusterTime)
 {
+	zend_error_handling   error_handling;
 	php_phongo_session_t* intern;
 	const bson_t*         cluster_time;
 	php_phongo_bson_state state;
@@ -182,9 +226,12 @@ static PHP_METHOD(Session, getClusterTime)
 	intern = Z_SESSION_OBJ_P(getThis());
 	SESSION_CHECK_LIVELINESS(intern, "getClusterTime")
 
+	zend_replace_error_handling(EH_THROW, phongo_exception_from_phongo_domain(PHONGO_ERROR_INVALID_ARGUMENT), &error_handling);
 	if (zend_parse_parameters_none() == FAILURE) {
+		zend_restore_error_handling(&error_handling);
 		return;
 	}
+	zend_restore_error_handling(&error_handling);
 
 	cluster_time = mongoc_client_session_get_cluster_time(intern->client_session);
 
@@ -198,17 +245,14 @@ static PHP_METHOD(Session, getClusterTime)
 		return;
 	}
 
-#if PHP_VERSION_ID >= 70000
 	RETURN_ZVAL(&state.zchild, 0, 1);
-#else
-	RETURN_ZVAL(state.zchild, 0, 1);
-#endif
 } /* }}} */
 
 /* {{{ proto object MongoDB\Driver\Session::getLogicalSessionId()
    Returns the logical session ID for this Session */
 static PHP_METHOD(Session, getLogicalSessionId)
 {
+	zend_error_handling   error_handling;
 	php_phongo_session_t* intern;
 	const bson_t*         lsid;
 	php_phongo_bson_state state;
@@ -218,9 +262,12 @@ static PHP_METHOD(Session, getLogicalSessionId)
 	intern = Z_SESSION_OBJ_P(getThis());
 	SESSION_CHECK_LIVELINESS(intern, "getLogicalSessionId")
 
+	zend_replace_error_handling(EH_THROW, phongo_exception_from_phongo_domain(PHONGO_ERROR_INVALID_ARGUMENT), &error_handling);
 	if (zend_parse_parameters_none() == FAILURE) {
+		zend_restore_error_handling(&error_handling);
 		return;
 	}
+	zend_restore_error_handling(&error_handling);
 
 	lsid = mongoc_client_session_get_lsid(intern->client_session);
 
@@ -230,26 +277,26 @@ static PHP_METHOD(Session, getLogicalSessionId)
 		return;
 	}
 
-#if PHP_VERSION_ID >= 70000
 	RETURN_ZVAL(&state.zchild, 0, 1);
-#else
-	RETURN_ZVAL(state.zchild, 0, 1);
-#endif
 } /* }}} */
 
 /* {{{ proto MongoDB\BSON\Timestamp|null MongoDB\Driver\Session::getOperationTime()
    Returns the operation time for this Session */
 static PHP_METHOD(Session, getOperationTime)
 {
+	zend_error_handling   error_handling;
 	php_phongo_session_t* intern;
 	uint32_t              timestamp, increment;
 
 	intern = Z_SESSION_OBJ_P(getThis());
 	SESSION_CHECK_LIVELINESS(intern, "getOperationTime")
 
+	zend_replace_error_handling(EH_THROW, phongo_exception_from_phongo_domain(PHONGO_ERROR_INVALID_ARGUMENT), &error_handling);
 	if (zend_parse_parameters_none() == FAILURE) {
+		zend_restore_error_handling(&error_handling);
 		return;
 	}
+	zend_restore_error_handling(&error_handling);
 
 	mongoc_client_session_get_operation_time(intern->client_session, &timestamp, &increment);
 
@@ -260,22 +307,26 @@ static PHP_METHOD(Session, getOperationTime)
 		RETURN_NULL();
 	}
 
-	php_phongo_bson_new_timestamp_from_increment_and_timestamp(return_value, increment, timestamp TSRMLS_CC);
+	php_phongo_bson_new_timestamp_from_increment_and_timestamp(return_value, increment, timestamp);
 } /* }}} */
 
 /* {{{ proto MongoDB\Driver\Server|null MongoDB\Driver\Session::getServer()
    Returns the server this session is pinned to */
 static PHP_METHOD(Session, getServer)
 {
+	zend_error_handling   error_handling;
 	php_phongo_session_t* intern;
 	uint32_t              server_id = 0;
 
 	intern = Z_SESSION_OBJ_P(getThis());
 	SESSION_CHECK_LIVELINESS(intern, "getServer")
 
+	zend_replace_error_handling(EH_THROW, phongo_exception_from_phongo_domain(PHONGO_ERROR_INVALID_ARGUMENT), &error_handling);
 	if (zend_parse_parameters_none() == FAILURE) {
+		zend_restore_error_handling(&error_handling);
 		return;
 	}
+	zend_restore_error_handling(&error_handling);
 
 	server_id = mongoc_client_session_get_server_id(intern->client_session);
 
@@ -284,118 +335,61 @@ static PHP_METHOD(Session, getServer)
 		RETURN_NULL();
 	}
 
-	phongo_server_init(return_value, intern->client, server_id TSRMLS_CC);
+	phongo_server_init(return_value, &intern->manager, server_id);
 } /* }}} */
 
 /* {{{ proto array|null MongoDB\Driver\Session::getTransactionOptions()
    Returns options for the currently running transaction */
 static PHP_METHOD(Session, getTransactionOptions)
 {
-	php_phongo_session_t*         intern;
-	mongoc_transaction_opt_t*     opts;
-	int64_t                       max_commit_time_ms;
-	const mongoc_read_concern_t*  read_concern;
-	const mongoc_read_prefs_t*    read_preference;
-	const mongoc_write_concern_t* write_concern;
+	zend_error_handling   error_handling;
+	php_phongo_session_t* intern;
 
 	intern = Z_SESSION_OBJ_P(getThis());
 	SESSION_CHECK_LIVELINESS(intern, "getTransactionOptions")
 
+	zend_replace_error_handling(EH_THROW, phongo_exception_from_phongo_domain(PHONGO_ERROR_INVALID_ARGUMENT), &error_handling);
 	if (zend_parse_parameters_none() == FAILURE) {
+		zend_restore_error_handling(&error_handling);
 		return;
 	}
+	zend_restore_error_handling(&error_handling);
 
-	opts = mongoc_session_opts_get_transaction_opts(intern->client_session);
-
-	if (!opts) {
-		return;
-	}
-
-	max_commit_time_ms = mongoc_transaction_opts_get_max_commit_time_ms(opts);
-	read_concern       = mongoc_transaction_opts_get_read_concern(opts);
-	read_preference    = mongoc_transaction_opts_get_read_prefs(opts);
-	write_concern      = mongoc_transaction_opts_get_write_concern(opts);
-
-	array_init_size(return_value, 4);
-
-	if (max_commit_time_ms) {
-		ADD_ASSOC_LONG_EX(return_value, "maxCommitTimeMS", max_commit_time_ms);
-	}
-
-	if (!mongoc_read_concern_is_default(read_concern)) {
-#if PHP_VERSION_ID >= 70000
-		zval zread_concern;
-
-		phongo_readconcern_init(&zread_concern, read_concern TSRMLS_CC);
-		ADD_ASSOC_ZVAL_EX(return_value, "readConcern", &zread_concern);
-#else
-		zval* zread_concern = NULL;
-		MAKE_STD_ZVAL(zread_concern);
-
-		phongo_readconcern_init(zread_concern, read_concern TSRMLS_CC);
-		ADD_ASSOC_ZVAL_EX(return_value, "readConcern", zread_concern);
-#endif
-	}
-
-	if (read_preference) {
-#if PHP_VERSION_ID >= 70000
-		zval zread_preference;
-
-		phongo_readpreference_init(&zread_preference, read_preference TSRMLS_CC);
-		ADD_ASSOC_ZVAL_EX(return_value, "readPreference", &zread_preference);
-#else
-		zval* zread_preference = NULL;
-		MAKE_STD_ZVAL(zread_preference);
-
-		phongo_readpreference_init(zread_preference, read_preference TSRMLS_CC);
-		ADD_ASSOC_ZVAL_EX(return_value, "readPreference", zread_preference);
-#endif
-	}
-
-	if (!mongoc_write_concern_is_default(write_concern)) {
-#if PHP_VERSION_ID >= 70000
-		zval zwrite_concern;
-
-		phongo_writeconcern_init(&zwrite_concern, write_concern TSRMLS_CC);
-		ADD_ASSOC_ZVAL_EX(return_value, "writeConcern", &zwrite_concern);
-#else
-		zval* zwrite_concern = NULL;
-		MAKE_STD_ZVAL(zwrite_concern);
-
-		phongo_writeconcern_init(zwrite_concern, write_concern TSRMLS_CC);
-		ADD_ASSOC_ZVAL_EX(return_value, "writeConcern", zwrite_concern);
-#endif
-	}
+	php_phongo_transaction_options_to_zval(intern->client_session, return_value);
 } /* }}} */
 
 /* {{{ proto string MongoDB\Driver\Session::getTransactionState()
    Returns the current transaction state for this session */
 static PHP_METHOD(Session, getTransactionState)
 {
+	zend_error_handling   error_handling;
 	php_phongo_session_t* intern;
 	const char*           state;
 
 	intern = Z_SESSION_OBJ_P(getThis());
 	SESSION_CHECK_LIVELINESS(intern, "getTransactionState")
 
+	zend_replace_error_handling(EH_THROW, phongo_exception_from_phongo_domain(PHONGO_ERROR_INVALID_ARGUMENT), &error_handling);
 	if (zend_parse_parameters_none() == FAILURE) {
+		zend_restore_error_handling(&error_handling);
 		return;
 	}
+	zend_restore_error_handling(&error_handling);
 
-	state = php_phongo_get_transaction_state_string(mongoc_client_session_get_transaction_state(intern->client_session) TSRMLS_CC);
+	state = php_phongo_get_transaction_state_string(mongoc_client_session_get_transaction_state(intern->client_session));
 	if (!state) {
 		/* Exception already thrown */
 		return;
 	}
 
-	PHONGO_RETURN_STRING(state);
+	RETURN_STRING(state);
 } /* }}} */
 
 /* Creates a opts structure from an array optionally containing an RP, RC,
  * WC object, and/or maxCommitTimeMS int. Returns NULL if no options were found,
  * or there was an invalid option. If there was an invalid option or structure,
  * an exception will be thrown too. */
-mongoc_transaction_opt_t* php_mongodb_session_parse_transaction_options(zval* options TSRMLS_DC)
+mongoc_transaction_opt_t* php_mongodb_session_parse_transaction_options(zval* options)
 {
 	mongoc_transaction_opt_t* opts = NULL;
 
@@ -403,7 +397,7 @@ mongoc_transaction_opt_t* php_mongodb_session_parse_transaction_options(zval* op
 		int64_t max_commit_time_ms = php_array_fetchc_long(options, "maxCommitTimeMS");
 
 		if (max_commit_time_ms < 0) {
-			phongo_throw_exception(PHONGO_ERROR_INVALID_ARGUMENT TSRMLS_CC, "Expected \"maxCommitTimeMS\" option to be >= 0, %" PRId64 " given", max_commit_time_ms);
+			phongo_throw_exception(PHONGO_ERROR_INVALID_ARGUMENT, "Expected \"maxCommitTimeMS\" option to be >= 0, %" PRId64 " given", max_commit_time_ms);
 			/* Freeing opts is not needed here, as it can't be set yet. The
 			 * code is here to keep it consistent with the others in case more
 			 * options are added before this one. */
@@ -414,7 +408,7 @@ mongoc_transaction_opt_t* php_mongodb_session_parse_transaction_options(zval* op
 		}
 
 		if (max_commit_time_ms > UINT32_MAX) {
-			phongo_throw_exception(PHONGO_ERROR_INVALID_ARGUMENT TSRMLS_CC, "Expected \"maxCommitTimeMS\" option to be <= %" PRIu32 ", %" PRId64 " given", UINT32_MAX, max_commit_time_ms);
+			phongo_throw_exception(PHONGO_ERROR_INVALID_ARGUMENT, "Expected \"maxCommitTimeMS\" option to be <= %" PRIu32 ", %" PRId64 " given", UINT32_MAX, max_commit_time_ms);
 			/* Freeing opts is not needed here, as it can't be set yet. The
 			 * code is here to keep it consistent with the others in case more
 			 * options are added before this one. */
@@ -434,8 +428,8 @@ mongoc_transaction_opt_t* php_mongodb_session_parse_transaction_options(zval* op
 	if (php_array_existsc(options, "readConcern")) {
 		zval* read_concern = php_array_fetchc(options, "readConcern");
 
-		if (Z_TYPE_P(read_concern) != IS_OBJECT || !instanceof_function(Z_OBJCE_P(read_concern), php_phongo_readconcern_ce TSRMLS_CC)) {
-			phongo_throw_exception(PHONGO_ERROR_INVALID_ARGUMENT TSRMLS_CC, "Expected \"readConcern\" option to be %s, %s given", ZSTR_VAL(php_phongo_readconcern_ce->name), PHONGO_ZVAL_CLASS_OR_TYPE_NAME_P(read_concern));
+		if (Z_TYPE_P(read_concern) != IS_OBJECT || !instanceof_function(Z_OBJCE_P(read_concern), php_phongo_readconcern_ce)) {
+			phongo_throw_exception(PHONGO_ERROR_INVALID_ARGUMENT, "Expected \"readConcern\" option to be %s, %s given", ZSTR_VAL(php_phongo_readconcern_ce->name), PHONGO_ZVAL_CLASS_OR_TYPE_NAME_P(read_concern));
 			if (opts) {
 				mongoc_transaction_opts_destroy(opts);
 			}
@@ -446,14 +440,14 @@ mongoc_transaction_opt_t* php_mongodb_session_parse_transaction_options(zval* op
 			opts = mongoc_transaction_opts_new();
 		}
 
-		mongoc_transaction_opts_set_read_concern(opts, phongo_read_concern_from_zval(read_concern TSRMLS_CC));
+		mongoc_transaction_opts_set_read_concern(opts, phongo_read_concern_from_zval(read_concern));
 	}
 
 	if (php_array_existsc(options, "readPreference")) {
 		zval* read_preference = php_array_fetchc(options, "readPreference");
 
-		if (Z_TYPE_P(read_preference) != IS_OBJECT || !instanceof_function(Z_OBJCE_P(read_preference), php_phongo_readpreference_ce TSRMLS_CC)) {
-			phongo_throw_exception(PHONGO_ERROR_INVALID_ARGUMENT TSRMLS_CC, "Expected \"readPreference\" option to be %s, %s given", ZSTR_VAL(php_phongo_readpreference_ce->name), PHONGO_ZVAL_CLASS_OR_TYPE_NAME_P(read_preference));
+		if (Z_TYPE_P(read_preference) != IS_OBJECT || !instanceof_function(Z_OBJCE_P(read_preference), php_phongo_readpreference_ce)) {
+			phongo_throw_exception(PHONGO_ERROR_INVALID_ARGUMENT, "Expected \"readPreference\" option to be %s, %s given", ZSTR_VAL(php_phongo_readpreference_ce->name), PHONGO_ZVAL_CLASS_OR_TYPE_NAME_P(read_preference));
 			if (opts) {
 				mongoc_transaction_opts_destroy(opts);
 			}
@@ -464,14 +458,14 @@ mongoc_transaction_opt_t* php_mongodb_session_parse_transaction_options(zval* op
 			opts = mongoc_transaction_opts_new();
 		}
 
-		mongoc_transaction_opts_set_read_prefs(opts, phongo_read_preference_from_zval(read_preference TSRMLS_CC));
+		mongoc_transaction_opts_set_read_prefs(opts, phongo_read_preference_from_zval(read_preference));
 	}
 
 	if (php_array_existsc(options, "writeConcern")) {
 		zval* write_concern = php_array_fetchc(options, "writeConcern");
 
-		if (Z_TYPE_P(write_concern) != IS_OBJECT || !instanceof_function(Z_OBJCE_P(write_concern), php_phongo_writeconcern_ce TSRMLS_CC)) {
-			phongo_throw_exception(PHONGO_ERROR_INVALID_ARGUMENT TSRMLS_CC, "Expected \"writeConcern\" option to be %s, %s given", ZSTR_VAL(php_phongo_writeconcern_ce->name), PHONGO_ZVAL_CLASS_OR_TYPE_NAME_P(write_concern));
+		if (Z_TYPE_P(write_concern) != IS_OBJECT || !instanceof_function(Z_OBJCE_P(write_concern), php_phongo_writeconcern_ce)) {
+			phongo_throw_exception(PHONGO_ERROR_INVALID_ARGUMENT, "Expected \"writeConcern\" option to be %s, %s given", ZSTR_VAL(php_phongo_writeconcern_ce->name), PHONGO_ZVAL_CLASS_OR_TYPE_NAME_P(write_concern));
 			if (opts) {
 				mongoc_transaction_opts_destroy(opts);
 			}
@@ -482,7 +476,7 @@ mongoc_transaction_opt_t* php_mongodb_session_parse_transaction_options(zval* op
 			opts = mongoc_transaction_opts_new();
 		}
 
-		mongoc_transaction_opts_set_write_concern(opts, phongo_write_concern_from_zval(write_concern TSRMLS_CC));
+		mongoc_transaction_opts_set_write_concern(opts, phongo_write_concern_from_zval(write_concern));
 	}
 
 	return opts;
@@ -492,6 +486,7 @@ mongoc_transaction_opt_t* php_mongodb_session_parse_transaction_options(zval* op
    Starts a new transaction */
 static PHP_METHOD(Session, startTransaction)
 {
+	zend_error_handling       error_handling;
 	php_phongo_session_t*     intern;
 	zval*                     options     = NULL;
 	mongoc_transaction_opt_t* txn_options = NULL;
@@ -500,19 +495,22 @@ static PHP_METHOD(Session, startTransaction)
 	intern = Z_SESSION_OBJ_P(getThis());
 	SESSION_CHECK_LIVELINESS(intern, "startTransaction")
 
-	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "|a!", &options) == FAILURE) {
+	zend_replace_error_handling(EH_THROW, phongo_exception_from_phongo_domain(PHONGO_ERROR_INVALID_ARGUMENT), &error_handling);
+	if (zend_parse_parameters(ZEND_NUM_ARGS(), "|a!", &options) == FAILURE) {
+		zend_restore_error_handling(&error_handling);
 		return;
 	}
+	zend_restore_error_handling(&error_handling);
 
 	if (options) {
-		txn_options = php_mongodb_session_parse_transaction_options(options TSRMLS_CC);
+		txn_options = php_mongodb_session_parse_transaction_options(options);
 	}
 	if (EG(exception)) {
 		return;
 	}
 
 	if (!mongoc_client_session_start_transaction(intern->client_session, txn_options, &error)) {
-		phongo_throw_exception_from_bson_error_t(&error TSRMLS_CC);
+		phongo_throw_exception_from_bson_error_t(&error);
 	}
 
 	if (txn_options) {
@@ -524,6 +522,7 @@ static PHP_METHOD(Session, startTransaction)
    Commits an existing transaction */
 static PHP_METHOD(Session, commitTransaction)
 {
+	zend_error_handling   error_handling;
 	php_phongo_session_t* intern;
 	bson_error_t          error;
 	bson_t                reply;
@@ -531,32 +530,40 @@ static PHP_METHOD(Session, commitTransaction)
 	intern = Z_SESSION_OBJ_P(getThis());
 	SESSION_CHECK_LIVELINESS(intern, "commitTransaction")
 
+	zend_replace_error_handling(EH_THROW, phongo_exception_from_phongo_domain(PHONGO_ERROR_INVALID_ARGUMENT), &error_handling);
 	if (zend_parse_parameters_none() == FAILURE) {
+		zend_restore_error_handling(&error_handling);
 		return;
 	}
+	zend_restore_error_handling(&error_handling);
 
 	if (!mongoc_client_session_commit_transaction(intern->client_session, &reply, &error)) {
-		phongo_throw_exception_from_bson_error_t_and_reply(&error, &reply TSRMLS_CC);
-		bson_destroy(&reply);
+		phongo_throw_exception_from_bson_error_t_and_reply(&error, &reply);
 	}
+
+	bson_destroy(&reply);
 } /* }}} */
 
 /* {{{ proto void MongoDB\Driver\Session::abortTransaction(void)
    Aborts (rolls back) an existing transaction */
 static PHP_METHOD(Session, abortTransaction)
 {
+	zend_error_handling   error_handling;
 	php_phongo_session_t* intern;
 	bson_error_t          error;
 
 	intern = Z_SESSION_OBJ_P(getThis());
 	SESSION_CHECK_LIVELINESS(intern, "abortTransaction")
 
+	zend_replace_error_handling(EH_THROW, phongo_exception_from_phongo_domain(PHONGO_ERROR_INVALID_ARGUMENT), &error_handling);
 	if (zend_parse_parameters_none() == FAILURE) {
+		zend_restore_error_handling(&error_handling);
 		return;
 	}
+	zend_restore_error_handling(&error_handling);
 
 	if (!mongoc_client_session_abort_transaction(intern->client_session, &error)) {
-		phongo_throw_exception_from_bson_error_t(&error TSRMLS_CC);
+		phongo_throw_exception_from_bson_error_t(&error);
 	}
 } /* }}} */
 
@@ -564,13 +571,17 @@ static PHP_METHOD(Session, abortTransaction)
    Ends the session, and a running transaction if active */
 static PHP_METHOD(Session, endSession)
 {
+	zend_error_handling   error_handling;
 	php_phongo_session_t* intern;
 
 	intern = Z_SESSION_OBJ_P(getThis());
 
+	zend_replace_error_handling(EH_THROW, phongo_exception_from_phongo_domain(PHONGO_ERROR_INVALID_ARGUMENT), &error_handling);
 	if (zend_parse_parameters_none() == FAILURE) {
+		zend_restore_error_handling(&error_handling);
 		return;
 	}
+	zend_restore_error_handling(&error_handling);
 
 	mongoc_client_session_destroy(intern->client_session);
 	intern->client_session = NULL;
@@ -580,14 +591,18 @@ static PHP_METHOD(Session, endSession)
    Returns whether a multi-document transaction is in progress */
 static PHP_METHOD(Session, isInTransaction)
 {
+	zend_error_handling   error_handling;
 	php_phongo_session_t* intern;
 
 	intern = Z_SESSION_OBJ_P(getThis());
 	SESSION_CHECK_LIVELINESS(intern, "isInTransaction")
 
+	zend_replace_error_handling(EH_THROW, phongo_exception_from_phongo_domain(PHONGO_ERROR_INVALID_ARGUMENT), &error_handling);
 	if (zend_parse_parameters_none() == FAILURE) {
+		zend_restore_error_handling(&error_handling);
 		return;
 	}
+	zend_restore_error_handling(&error_handling);
 
 	RETURN_BOOL(mongoc_client_session_in_transaction(intern->client_session));
 } /* }}} */
@@ -633,62 +648,51 @@ static zend_function_entry php_phongo_session_me[] = {
 /* {{{ MongoDB\Driver\Session object handlers */
 static zend_object_handlers php_phongo_handler_session;
 
-static void php_phongo_session_free_object(phongo_free_object_arg* object TSRMLS_DC) /* {{{ */
+static void php_phongo_session_free_object(zend_object* object) /* {{{ */
 {
 	php_phongo_session_t* intern = Z_OBJ_SESSION(object);
 
-	zend_object_std_dtor(&intern->std TSRMLS_CC);
+	zend_object_std_dtor(&intern->std);
 
 	/* If this Session was created in a different process, reset the client so
 	 * that its session pool is cleared and mongoc_client_session_destroy will
 	 * destroy the corresponding server session rather than return it to the
 	 * now-empty pool. This will ensure that we do not re-use a server session
 	 * (i.e. LSID) created by a parent process. */
-	PHONGO_RESET_CLIENT_IF_PID_DIFFERS(intern);
+	PHONGO_RESET_CLIENT_IF_PID_DIFFERS(intern, Z_MANAGER_OBJ_P(&intern->manager));
 
 	if (intern->client_session) {
 		mongoc_client_session_destroy(intern->client_session);
 	}
 
-#if PHP_VERSION_ID < 70000
-	efree(intern);
-#endif
+	if (!Z_ISUNDEF(intern->manager)) {
+		zval_ptr_dtor(&intern->manager);
+	}
 } /* }}} */
 
-static phongo_create_object_retval php_phongo_session_create_object(zend_class_entry* class_type TSRMLS_DC) /* {{{ */
+static zend_object* php_phongo_session_create_object(zend_class_entry* class_type) /* {{{ */
 {
 	php_phongo_session_t* intern = NULL;
 
 	intern = PHONGO_ALLOC_OBJECT_T(php_phongo_session_t, class_type);
 
-	zend_object_std_init(&intern->std, class_type TSRMLS_CC);
+	zend_object_std_init(&intern->std, class_type);
 	object_properties_init(&intern->std, class_type);
 
 	PHONGO_SET_CREATED_BY_PID(intern);
 
-#if PHP_VERSION_ID >= 70000
 	intern->std.handlers = &php_phongo_handler_session;
 
 	return &intern->std;
-#else
-	{
-		zend_object_value retval;
-		retval.handle   = zend_objects_store_put(intern, (zend_objects_store_dtor_t) zend_objects_destroy_object, php_phongo_session_free_object, NULL TSRMLS_CC);
-		retval.handlers = &php_phongo_handler_session;
-
-		return retval;
-	}
-#endif
 } /* }}} */
 
-static HashTable* php_phongo_session_get_debug_info(zval* object, int* is_temp TSRMLS_DC) /* {{{ */
+static HashTable* php_phongo_session_get_debug_info(phongo_compat_object_handler_type* object, int* is_temp) /* {{{ */
 {
-	php_phongo_session_t*       intern = NULL;
-	const mongoc_session_opt_t* cs_opts;
-	zval                        retval = ZVAL_STATIC_INIT;
+	php_phongo_session_t* intern = NULL;
+	zval                  retval = ZVAL_STATIC_INIT;
 
 	*is_temp = 1;
-	intern   = Z_SESSION_OBJ_P(object);
+	intern   = Z_OBJ_SESSION(PHONGO_COMPAT_GET_OBJ(object));
 
 	array_init(&retval);
 
@@ -705,11 +709,7 @@ static HashTable* php_phongo_session_get_debug_info(zval* object, int* is_temp T
 			goto done;
 		}
 
-#if PHP_VERSION_ID >= 70000
 		ADD_ASSOC_ZVAL_EX(&retval, "logicalSessionId", &state.zchild);
-#else
-		ADD_ASSOC_ZVAL_EX(&retval, "logicalSessionId", state.zchild);
-#endif
 	} else {
 		ADD_ASSOC_NULL_EX(&retval, "logicalSessionId");
 	}
@@ -729,11 +729,7 @@ static HashTable* php_phongo_session_get_debug_info(zval* object, int* is_temp T
 				goto done;
 			}
 
-#if PHP_VERSION_ID >= 70000
 			ADD_ASSOC_ZVAL_EX(&retval, "clusterTime", &state.zchild);
-#else
-			ADD_ASSOC_ZVAL_EX(&retval, "clusterTime", state.zchild);
-#endif
 		} else {
 			ADD_ASSOC_NULL_EX(&retval, "clusterTime");
 		}
@@ -742,10 +738,12 @@ static HashTable* php_phongo_session_get_debug_info(zval* object, int* is_temp T
 	}
 
 	if (intern->client_session) {
-		cs_opts = mongoc_client_session_get_opts(intern->client_session);
+		const mongoc_session_opt_t* cs_opts = mongoc_client_session_get_opts(intern->client_session);
 		ADD_ASSOC_BOOL_EX(&retval, "causalConsistency", mongoc_session_opts_get_causal_consistency(cs_opts));
+		ADD_ASSOC_BOOL_EX(&retval, "snapshot", mongoc_session_opts_get_snapshot(cs_opts));
 	} else {
 		ADD_ASSOC_NULL_EX(&retval, "causalConsistency");
+		ADD_ASSOC_NULL_EX(&retval, "snapshot");
 	}
 
 	if (intern->client_session) {
@@ -754,18 +752,10 @@ static HashTable* php_phongo_session_get_debug_info(zval* object, int* is_temp T
 		mongoc_client_session_get_operation_time(intern->client_session, &timestamp, &increment);
 
 		if (timestamp && increment) {
-#if PHP_VERSION_ID >= 70000
 			zval ztimestamp;
 
-			php_phongo_bson_new_timestamp_from_increment_and_timestamp(&ztimestamp, increment, timestamp TSRMLS_CC);
+			php_phongo_bson_new_timestamp_from_increment_and_timestamp(&ztimestamp, increment, timestamp);
 			ADD_ASSOC_ZVAL_EX(&retval, "operationTime", &ztimestamp);
-#else
-			zval* ztimestamp;
-
-			MAKE_STD_ZVAL(ztimestamp);
-			php_phongo_bson_new_timestamp_from_increment_and_timestamp(ztimestamp, increment, timestamp TSRMLS_CC);
-			ADD_ASSOC_ZVAL_EX(&retval, "operationTime", ztimestamp);
-#endif
 		} else {
 			ADD_ASSOC_NULL_EX(&retval, "operationTime");
 		}
@@ -778,18 +768,10 @@ static HashTable* php_phongo_session_get_debug_info(zval* object, int* is_temp T
 
 		if (server_id) {
 
-#if PHP_VERSION_ID >= 70000
 			zval server;
 
-			phongo_server_init(&server, intern->client, server_id TSRMLS_CC);
+			phongo_server_init(&server, &intern->manager, server_id);
 			ADD_ASSOC_ZVAL_EX(&retval, "server", &server);
-#else
-			zval* server = NULL;
-
-			MAKE_STD_ZVAL(server);
-			phongo_server_init(server, intern->client, server_id TSRMLS_CC);
-			ADD_ASSOC_ZVAL_EX(&retval, "server", server);
-#endif
 		} else {
 			ADD_ASSOC_NULL_EX(&retval, "server");
 		}
@@ -797,6 +779,33 @@ static HashTable* php_phongo_session_get_debug_info(zval* object, int* is_temp T
 		ADD_ASSOC_NULL_EX(&retval, "server");
 	}
 
+	if (intern->client_session) {
+		ADD_ASSOC_BOOL_EX(&retval, "inTransaction", mongoc_client_session_in_transaction(intern->client_session));
+	} else {
+		ADD_ASSOC_NULL_EX(&retval, "inTransaction");
+	}
+
+	if (intern->client_session) {
+		const char* state = php_phongo_get_transaction_state_string(mongoc_client_session_get_transaction_state(intern->client_session));
+
+		if (!state) {
+			/* Exception should already have been thrown */
+			goto done;
+		}
+
+		ADD_ASSOC_STRING(&retval, "transactionState", state);
+	} else {
+		ADD_ASSOC_NULL_EX(&retval, "transactionState");
+	}
+
+	if (intern->client_session) {
+		zval txn_opts;
+		php_phongo_transaction_options_to_zval(intern->client_session, &txn_opts);
+		ADD_ASSOC_ZVAL_EX(&retval, "transactionOptions", &txn_opts);
+	} else {
+		ADD_ASSOC_NULL_EX(&retval, "transactionOptions");
+	}
+
 done:
 	return Z_ARRVAL(retval);
 } /* }}} */
@@ -807,23 +816,21 @@ void php_phongo_session_init_ce(INIT_FUNC_ARGS) /* {{{ */
 	zend_class_entry ce;
 
 	INIT_NS_CLASS_ENTRY(ce, "MongoDB\\Driver", "Session", php_phongo_session_me);
-	php_phongo_session_ce                = zend_register_internal_class(&ce TSRMLS_CC);
+	php_phongo_session_ce                = zend_register_internal_class(&ce);
 	php_phongo_session_ce->create_object = php_phongo_session_create_object;
 	PHONGO_CE_FINAL(php_phongo_session_ce);
 	PHONGO_CE_DISABLE_SERIALIZATION(php_phongo_session_ce);
 
 	memcpy(&php_phongo_handler_session, phongo_get_std_object_handlers(), sizeof(zend_object_handlers));
 	php_phongo_handler_session.get_debug_info = php_phongo_session_get_debug_info;
-#if PHP_VERSION_ID >= 70000
-	php_phongo_handler_session.free_obj = php_phongo_session_free_object;
-	php_phongo_handler_session.offset   = XtOffsetOf(php_phongo_session_t, std);
-#endif
+	php_phongo_handler_session.free_obj       = php_phongo_session_free_object;
+	php_phongo_handler_session.offset         = XtOffsetOf(php_phongo_session_t, std);
 
-	zend_declare_class_constant_string(php_phongo_session_ce, ZEND_STRL("TRANSACTION_NONE"), PHONGO_TRANSACTION_NONE TSRMLS_CC);
-	zend_declare_class_constant_string(php_phongo_session_ce, ZEND_STRL("TRANSACTION_STARTING"), PHONGO_TRANSACTION_STARTING TSRMLS_CC);
-	zend_declare_class_constant_string(php_phongo_session_ce, ZEND_STRL("TRANSACTION_IN_PROGRESS"), PHONGO_TRANSACTION_IN_PROGRESS TSRMLS_CC);
-	zend_declare_class_constant_string(php_phongo_session_ce, ZEND_STRL("TRANSACTION_COMMITTED"), PHONGO_TRANSACTION_COMMITTED TSRMLS_CC);
-	zend_declare_class_constant_string(php_phongo_session_ce, ZEND_STRL("TRANSACTION_ABORTED"), PHONGO_TRANSACTION_ABORTED TSRMLS_CC);
+	zend_declare_class_constant_string(php_phongo_session_ce, ZEND_STRL("TRANSACTION_NONE"), PHONGO_TRANSACTION_NONE);
+	zend_declare_class_constant_string(php_phongo_session_ce, ZEND_STRL("TRANSACTION_STARTING"), PHONGO_TRANSACTION_STARTING);
+	zend_declare_class_constant_string(php_phongo_session_ce, ZEND_STRL("TRANSACTION_IN_PROGRESS"), PHONGO_TRANSACTION_IN_PROGRESS);
+	zend_declare_class_constant_string(php_phongo_session_ce, ZEND_STRL("TRANSACTION_COMMITTED"), PHONGO_TRANSACTION_COMMITTED);
+	zend_declare_class_constant_string(php_phongo_session_ce, ZEND_STRL("TRANSACTION_ABORTED"), PHONGO_TRANSACTION_ABORTED);
 } /* }}} */
 
 /*
